package org.apithefire.util.lang;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Comparator;
import java.util.List;

public class Path implements Comparable<Path> {
	
	private static class PathComparator implements Comparator<Path> {
		
		private final Comparator<String> componentComparator;
		
		public PathComparator() {
			componentComparator = new Comparator<String>() {
				public int compare(String string1, String string2) {
					return string1.compareTo(string2);
				}
			};
		}
		
		public PathComparator(Comparator<String> componentComparator) {
			this.componentComparator = Objects.nonNull(componentComparator); 
		}

		public int compare(Path path1, Path path2) {
			for (int i = 0; i < path1.components.length
					&& i < path2.components.length; ++i) {
				
				int diff = componentComparator.compare(path1.components[i],
						path2.components[i]);

				if (diff != 0) {
					return diff;
				}
			}
			return path1.components.length - path2.components.length;
		}
		
	}
	
	public static final Path EMPTY_PATH = new Path();
	
	private static final String DEFAULT_SEPARATOR = "/";
	private static final PathComparator DEFAULT_COMPARATOR = new PathComparator();
	
	private final String[] components;
	private String defaultString;
	
	public Path() {
		components = new String[0];
	}
	
	private Path(int length) {
		components = new String[length];
	}
	
	public Path(String component) {
		components = new String[] { component };
	}
	
	public Path(String... components) {
		this.components = new String[components.length];
		
		for (int i = 0; i < this.components.length; ++i) {
			this.components[i] = Objects.nonNull(components[i]);
		}
	}
	
	public Path(List<String> components) {
		this.components = new String[components.size()];

		for (int i = 0; i < this.components.length; ++i) {
			this.components[i] = Objects.nonNull(components.get(i));
		}
	}
	
	private static void throwEmptySeparatorArgument() {
		throw new IllegalArgumentException("Path separator is empty.");
	}
	
	public static Path split(String path) {
		return split(path, DEFAULT_SEPARATOR);
	}
	
	public static Path split(String string, String separator) {
		if (separator.length() == 0) {
			throwEmptySeparatorArgument();
		}
		
		int start = 0;
		int end = string.indexOf(separator);
		List<String> components = new ArrayList<String>();
		
		String component = null;
	
		while (end != -1) {
			component = string.substring(start, end);
			components.add(component);
			start = end + separator.length();
			end = string.indexOf(separator, start);
		}
		component = string.substring(start);
		components.add(component);
		return new Path(components);
	}
	
	public boolean isEmpty() {
		return components.length < 1;
	}
	
	public int length() {
		return components.length;
	}
	
	public String componentAt(int index) {
		return components[index];
	}
	
	public Path trim() {
		return trim(true, true);
	}
	
	private Path trim(boolean trimLeft, boolean trimRight) {
		if (components.length == 0) {
			return new Path();
		}
		
		int left = 0;
		int right = components.length;
		
		if (trimLeft) {
			while (left < components.length && components[left].length() == 0) {
				++left;
			}
		}
		if (trimRight) {
			while (right > 0 && components[right - 1].length() == 0) {
				--right;
			}
		}
		if (left > right) {
			return new Path();
		}
		
		Path path = new Path(right - left);
		
		for (int i = left; i < right; ++i) {
			path.components[i - left] = components[i];
		}
		return path;
	}
	
	public Path trimLeft() {
		return trim(true, false);
	}
	
	public Path trimRight() {
		return trim(false, true);
	}
	
	public Path concat(Path path) {
		int concatLength = components.length + path.components.length;
		Path concatPath = new Path(concatLength);
		
		for (int i = 0; i < components.length; ++i) {
			concatPath.components[i] = components[i];
		}
		for (int i = 0; i < path.components.length; ++i) {
			concatPath.components[components.length + i] = path.components[i];
		}
		return concatPath;
	}
	
	public Path subPath(int fromIndex) {
		return subPath(fromIndex, components.length);
	}
	
	public Path subPath(int beginIndex, int endIndex) {
		if (beginIndex > endIndex) {
			throw new IllegalArgumentException(
					"beginIndex is greater than endIndex.");
		}
		
		Path subPath = new Path(endIndex - beginIndex);
		
		for (int i = beginIndex; i < endIndex; ++i) {
			subPath.components[i - beginIndex] = components[i];
		}
		return subPath;
	}
	
	public int indexOf(String component) {
		Objects.nonNull(component);
		
		for (int i = 0; i < components.length; ++i) {
			if (components[i].equals(component)) {
				return i;
			}
		}
		return -1;
	}
	
	public int indexOf(Path path) {
		return indexOf(components, path.components);
	}
	
	private static int indexOf(String[] c1, String[] c2) {
		if (c1.length < c2.length) {
			return -1;
		}
		if (c1.length == 0 && c2.length == 0) {
			return 0;
		}
		for (int i = 0; i < c1.length - c2.length + 1; ++i) {
			if (startsWith(c1, i, c2)) {
				return i;
			}
		}
		return -1;
	}
	
	protected static boolean startsWith(String[] c1, int c1Begin, String[] c2) {
		for (int c1Index = c1Begin; c1Index < c1.length; ++c1Index) {
			int c2Index = 0;
			
			while (c2Index < c2.length
					&& c1Index + c2Index < c1.length
					&& c1[c1Index + c2Index].equals(c2[c2Index])) {
				++c2Index;
			}
			return c2Index == c2.length;
		}
		return c2.length == 0;
	}
	
	public boolean startsWith(String component) {
		return indexOf(component) == 0;
	}
	
	public boolean startsWith(Path path) {
		return indexOf(path) == 0;
	}
	
	public int lastIndexOf(String component) {
		Objects.nonNull(component);
		
		for (int i = components.length - 1; i >= 0; --i) {
			if (components[i].equals(component)) {
				return i;
			}
		}
		return -1;
	}
	
	public int lastIndexOf(Path path) {
		return lastIndexOf(components, path.components);
	}
	
	private static int lastIndexOf(String[] c1, String[] c2) {
		if (c1.length < c2.length) {
			return -1;
		}
		if (c1.length == 0 && c2.length == 0) {
			return 0;
		}
		for (int i = c1.length - c2.length; i >= 0; --i) {
			if (startsWith(c1, i, c2)) {
				return i;
			}
		}
		return -1;
	}
	
	public boolean endsWith(String component) {
		int index = lastIndexOf(component);
		
		return index == -1 ? false : index == components.length - 1;
	}
	
	public boolean endsWith(Path path) {
		int index = lastIndexOf(path);
		
		if (index == -1) {
			return false;
		}
		return index == components.length - path.components.length;
	}
	
	public boolean contains(Path path) {
		return indexOf(path) != -1;
	}
	
	public boolean contains(String component) {
		return indexOf(component) != -1;
	}
	
	public int compareTo(Path path) {
		return DEFAULT_COMPARATOR.compare(this, path);
	}
	
	public static Comparator<Path> createComprator() {
		return new PathComparator();
	}
	
	public static Comparator<Path> createComparator(
			Comparator<String> componentComparator) {
		return new PathComparator(componentComparator);
	}
	
	@Override
	public int hashCode() {
		return toString().hashCode();
	}
	
	@Override
	public boolean equals(Object object) {
		if (!(object instanceof Path)) {
			return false;
		}
		return Arrays.equals(components, ((Path) object).components);
	}
	
	public String toString(String separator) {
		if (separator.length() == 0) {
			throwEmptySeparatorArgument();
		}
		if (components.length == 0) {
			return "";
		}
		
		StringBuilder builder = new StringBuilder();
		int componentCount = components.length;
		
		for (int i = 0; i < componentCount; ++i) {
			String component = components[i];
			
			if (i > 0 && componentCount > 1) {
				builder.append(separator);
			}
			builder.append(component);
		}
		return builder.toString();
	}
	
	@Override
	public String toString() {
		if (defaultString == null) {
			defaultString = toString(DEFAULT_SEPARATOR);
		}
		return defaultString;
	}

}
