package org.apithefire.util.resource;

import java.io.InputStream;
import java.util.ArrayList;
import java.util.Comparator;
import java.util.Date;
import java.util.Enumeration;
import java.util.HashSet;
import java.util.List;
import java.util.Set;
import java.util.SortedMap;
import java.util.TreeMap;
import java.util.jar.JarEntry;
import java.util.jar.JarFile;

import org.apithefire.util.lang.Objects;
import org.apithefire.util.lang.Path;
import org.apithefire.util.runtimeexception.io.RuntimeIoException;

public class JarResource extends AbstractResource {
	
	protected static class EmptyIsMaxStringComparator implements Comparator<String> {
		public int compare(String string1, String string2) {
			int length1 = string1.length();
			int length2 = string2.length();
			
			if (length1 == 0 || length2 == 0) {
				if (length1 == length2) {
					return 0;
				}
				return length1 == 0 ? 1 : -1;
			}
			return string1.compareTo(string2);
		}	
	}
	
	private static final Comparator<Path> PATH_COMPARATOR
		= Path.createComparator(new EmptyIsMaxStringComparator());
	
	private final JarFile jarFile;
	private final SortedMap<Path, JarEntry> jarEntryMap;
	private final Path path;
	
	public JarResource(JarFile file) {
		this(file, jarEntryMap(file), Path.EMPTY_PATH);
	}
	
	private static SortedMap<Path, JarEntry> jarEntryMap(JarFile file) {
		SortedMap<Path, JarEntry> entryMap = new TreeMap<Path, JarEntry>(
				PATH_COMPARATOR);
		Enumeration<JarEntry> entryEnm = file.entries();
		
		while (entryEnm.hasMoreElements()) {
			JarEntry iterEntry = entryEnm.nextElement();
			Path iterPath = Path.split(iterEntry.getName()).trim();
			
			entryMap.put(iterPath, iterEntry);
		}
		return entryMap;
	}
	
	private JarResource(JarFile jarFile, SortedMap<Path, JarEntry> jarEntryMap,
			Path path) {
		this.jarFile = Objects.nonNull(jarFile);
		this.jarEntryMap = jarEntryMap;
		this.path = path;
	}
	
	public boolean isDirectory() {
		return true;
	}

	public List<String> listResourceNames(NameFilter filter) {
		Set<String> resourceNames = new HashSet<String>();
		
		for (Path jarEntryPath : jarEntryMap.keySet()) {
			if (path.length() < jarEntryPath.length()) {
				String name = jarEntryPath.componentAt(path.length());
				String resourceName = filter.filter(name);
				
				if (resourceName != null && resourceName.length() > 0) {
					resourceNames.add(resourceName);
				}
			}
		}
		return new ArrayList<String>(resourceNames);
	}

	@Override
	protected Resource doGetResource(Path relativePath) {
		Path fromPath = this.path.concat(relativePath);
		Path toPath = fromPath.concat(new Path(""));

		SortedMap<Path, JarEntry> subJarEntryMap = jarEntryMap.subMap(
				fromPath, toPath);
		
		JarEntry contentJarEntry = subJarEntryMap.get(fromPath);
		
		if (contentJarEntry != null
				&& !contentJarEntry.isDirectory()
				&& subJarEntryMap.size() == 1) {
			return new ContentJarResource(jarFile, contentJarEntry);
		}
		if (subJarEntryMap.isEmpty()) {
			throw new ResourceNotFoundException();
		}
		return new JarResource(jarFile, subJarEntryMap, fromPath);
	}
	
	public boolean isContent() {
		return false;
	}

	public Date getContentModifiedTime() {
		// XXX: shouldn't we return null?
		return new Date(0);
	}

	public long getContentSize() {
		return 0;
	}

	public InputStream openContent() {
		throw new RuntimeIoException("Can not open content for "
				+ toString() + ".");
	}
	
	@Override
	public String toString() {
		return Objects.toString(this)
				.property("jarFile", jarFile)
				.property("path", path)
				.toString();
	}

}
